import * as SQLite from 'wa-sqlite';
import SQLiteAsyncModule from 'wa-sqlite/dist/wa-sqlite-async.mjs';
import * as VFS from 'wa-sqlite/src/VFS.js';

const RETRYABLE_EXCEPTIONS = new Set([
    "TransactionInactiveError",
    "InvalidStateError",
]);
// For debugging.
let nextTxId = 0;
const mapTxToId = new WeakMap();
function log$3(...args) {
    // console.debug(...args);
}
// This class manages IDBTransaction and IDBRequest instances. It tries
// to reuse transactions to minimize transaction overhead.
class IDBContext {
    dbReady;
    txOptions;
    tx;
    txComplete;
    request;
    // eslint-disable-next-line @typescript-eslint/no-explicit-any
    chain = Promise.resolve();
    constructor(idbDatabase, txOptions = { durability: "default" }) {
        this.dbReady = Promise.resolve(idbDatabase);
        this.txOptions = txOptions;
    }
    /**
     * Run a function with the provided object stores. The function
     * should be idempotent in case it is passed an expired transaction.
     */
    async run(mode, f) {
        // Ensure that functions run sequentially.
        return (this.chain = this.chain.then(() => {
            return this.performRun(mode, f);
        }));
    }
    async performRun(mode, f) {
        const db = await this.dbReady;
        const storeNames = Array.from(db.objectStoreNames);
        if (mode !== "readonly" && this.tx?.mode === "readonly") {
            // Force creation of a new read-write transaction.
            this.tx = undefined;
        }
        else if (this.request?.readyState === "pending") {
            const req = this.request;
            // Wait for pending IDBRequest so the IDBTransaction is active.
            await new Promise((resolve) => {
                req.addEventListener("success", resolve);
                req.addEventListener("error", resolve);
            });
        }
        let res = "not_set";
        // Run the user function with a retry in case the transaction is invalid.
        for (let i = 0; i < 2; ++i) {
            if (!this.tx) {
                // @ts-expect-error lib dom misses third argument
                this.tx = db.transaction(storeNames, mode, this.txOptions);
                // const timeStart = Date.now();
                // console.log("start transaction");
                const tx = this.tx;
                this.txComplete = new Promise((resolve) => {
                    tx.addEventListener("complete", () => {
                        if (this.tx === tx) {
                            this.tx = undefined;
                        }
                        resolve();
                        // console.log("end transaction", Date.now() - timeStart);
                        log$3(`transaction ${mapTxToId.get(tx)} complete`);
                    });
                });
                mapTxToId.set(this.tx, nextTxId++);
            }
            try {
                const tx = this.tx;
                const stores = Object.fromEntries(storeNames.map((name) => {
                    const objectStore = tx.objectStore(name);
                    const store = new Store(objectStore, (request) => this.setRequest(request));
                    return [name, store];
                }));
                // console.log(stores, storeNames);
                res = await f(stores);
            }
            catch (e) {
                if (i || !RETRYABLE_EXCEPTIONS.has(e.name)) {
                    // On failure make sure nothing is committed.
                    try {
                        this.tx.abort();
                        // eslint-disable-next-line no-empty
                    }
                    catch (ignored) { }
                    throw e;
                }
                this.tx = undefined;
            }
        }
        if (res === "not_set") {
            throw new Error("No result returns, that should not happen!");
        }
        return res;
    }
    async sync() {
        const request = this.request;
        if (request?.readyState === "pending") {
            await new Promise((resolve) => {
                request.addEventListener("success", resolve);
                request.addEventListener("error", resolve);
            });
            request.transaction?.commit();
        }
        return this.txComplete;
    }
    setRequest(request) {
        this.request = request;
        return new Promise((resolve, reject) => {
            request.onsuccess = () => {
                // console.log(request.result);
                resolve(request.result);
            };
            request.onerror = () => reject(request.error);
        });
    }
}
// const cache = new Map<number, Int8Array>();
// IDBStore wrapper passed to IDBActivity run functions.
class Store {
    store;
    // eslint-disable-next-line @typescript-eslint/no-explicit-any
    addRequest;
    constructor(store, 
    // eslint-disable-next-line @typescript-eslint/no-explicit-any
    addRequest) {
        this.store = store;
        this.addRequest = addRequest;
    }
    get(query) {
        // if (cache.has(query.lower[1])) {
        //   console.log("cache hit!");
        //   return this.addRequest({
        //     set onsuccess(call) {
        //       call();
        //     },
        //     onerror: () => null,
        //     result: { data: cache.get(query.lower[1]) },
        //   });
        // }
        // log(`get ${this.store.name}`, query.lower);
        const request = this.store.get(query);
        return this.addRequest(request);
    }
    getAll(query, count) {
        log$3(`getAll ${this.store.name}`, query, count);
        const request = this.store.getAll(query, count);
        return this.addRequest(request);
    }
    getKey(query) {
        log$3(`getKey ${this.store.name}`, query);
        const request = this.store.getKey(query);
        return this.addRequest(request);
    }
    getAllKeys(query, count) {
        log$3(`getAllKeys ${this.store.name}`, query, count);
        const request = this.store.getAllKeys(query, count);
        return this.addRequest(request);
    }
    put(value, key) {
        // cache.set(value.offset, value.data);
        log$3(`put ${this.store.name}`, value, key);
        const request = this.store.put(value, key);
        return this.addRequest(request);
    }
    delete(query) {
        console.trace();
        log$3(`delete ${this.store.name}`, query);
        const request = this.store.delete(query);
        return this.addRequest(request);
    }
    clear() {
        log$3(`clear ${this.store.name}`);
        const request = this.store.clear();
        return this.addRequest(request);
    }
    index(name) {
        return new Index(this.store.index(name), (request) => this.addRequest(request));
    }
}
class Index {
    index;
    // eslint-disable-next-line @typescript-eslint/no-explicit-any
    addRequest;
    constructor(index, 
    // eslint-disable-next-line @typescript-eslint/no-explicit-any
    addRequest) {
        this.index = index;
        this.addRequest = addRequest;
    }
    getAllKeys(query, count) {
        log$3(`IDBIndex.getAllKeys ${this.index.objectStore.name}<${this.index.name}>`, query, count);
        const request = this.index.getAllKeys(query, count);
        return this.addRequest(request);
    }
}

const WEB_LOCKS = 
// @ts-ignore
navigator["locks"] ??
    console.warn("concurrency is unsafe without Web Locks API");
class WebLocks {
    mapIdToState = new Map();
    mapNameToReleaser = new Map();
    // Use a single exclusive lock.
    async lock(name, flags) {
        const lockState = this.mapIdToState.get(name) ?? VFS.SQLITE_LOCK_NONE;
        if (lockState === VFS.SQLITE_LOCK_NONE) {
            await this.acquireWebLock(name, "exclusive");
        }
        this.mapIdToState.set(name, flags);
        return VFS.SQLITE_OK;
    }
    async unlock(name, flags) {
        if (flags === VFS.SQLITE_LOCK_NONE) {
            this.releaseWebLock(name);
            this.mapIdToState.delete(name);
        }
        else {
            this.mapIdToState.set(name, flags);
        }
        return VFS.SQLITE_OK;
    }
    async acquireWebLock(name, mode, signal) {
        if (WEB_LOCKS) {
            const lockName = `sqlite-${name}`;
            // eslint-disable-next-line promise/param-names, no-async-promise-executor
            return new Promise(async (hasLock, aborted) => {
                try {
                    await WEB_LOCKS.request(lockName, { mode, signal }, () => 
                    // eslint-disable-next-line promise/param-names
                    new Promise((release) => {
                        this.mapNameToReleaser.set(name, release);
                        hasLock();
                    }));
                }
                catch (e) {
                    aborted(e);
                }
            });
        }
    }
    releaseWebLock(name) {
        this.mapNameToReleaser.get(name)?.();
        this.mapNameToReleaser.delete(name);
    }
}

const SECTOR_SIZE$1 = 512;
const blockSize = 32 * 1024;
const DEFAULT_OPTIONS$2 = {
    durability: "default",
    purge: "deferred",
    purgeAtLeast: 16,
};
function log$2(...args) {
    // console.debug(...args);
}
class IDBAtomicVFS extends VFS.Base {
    name;
    options;
    filesState = new Map();
    dbNameFileIdMap = new Map();
    webLocks = new WebLocks();
    pendingPurges = new Set();
    blockToWrite = new Map();
    constructor(name, options = DEFAULT_OPTIONS$2) {
        super();
        this.name = name;
        this.options = { ...DEFAULT_OPTIONS$2, ...options };
    }
    xOpen(name, fileId, flags, pOutFlags) {
        console.warn("...opening", name);
        return this.handleAsync(async () => {
            this.dbNameFileIdMap.set(name, fileId);
            const db = new IDBContext(openDatabase$2(name), {
                durability: this.options.durability,
            });
            if (name === null)
                name = `null_${fileId}`;
            log$2(`xOpen ${name} 0x${fileId.toString(16)} 0x${flags.toString(16)}`);
            try {
                const file = {
                    name,
                    flags,
                    block0: undefined,
                };
                this.filesState.set(fileId, {
                    file,
                    db,
                });
                // Read the first block, which also contains the file metadata.
                file.block0 = await db.run("readonly", ({ blocks }) => blocks.get(this.#bound(0)));
                if (!file.block0) {
                    // File doesn't exist, create if requested.
                    if (flags & VFS.SQLITE_OPEN_CREATE) {
                        file.block0 = {
                            offset: 0,
                            data: new Int8Array(0),
                            fileSize: 0,
                        };
                        // Write metadata block to IndexedDB.
                        await db.run("readwrite", ({ blocks }) => blocks.put(file.block0, 0));
                        await db.sync();
                    }
                    else {
                        throw new Error(`file not found: ${name}`);
                    }
                }
                pOutFlags.set(flags & VFS.SQLITE_OPEN_READONLY);
                return VFS.SQLITE_OK;
            }
            catch (e) {
                console.error(e);
                return VFS.SQLITE_CANTOPEN;
            }
        });
    }
    xClose(fileId) {
        return this.handleAsync(async () => {
            try {
                const state = this.filesState.get(fileId);
                if (state) {
                    const { file } = state;
                    log$2(`xClose ${file.name}`);
                    this.filesState.delete(fileId);
                    if (file.flags & VFS.SQLITE_OPEN_DELETEONCLOSE) {
                        // TODO: delete DB
                    }
                }
                return VFS.SQLITE_OK;
            }
            catch (e) {
                console.error(e);
                return VFS.SQLITE_IOERR;
            }
        });
    }
    xRead(fileId, pData, iOffset) {
        return this.handleAsync(async () => {
            const { file, db } = this.getFileStateByIdOrThrow(fileId);
            log$2(`xRead ${file.name} ${pData.value.length} ${iOffset}`);
            const fromCache = this.blockToWrite.get(-iOffset)?.data;
            if (fromCache && fromCache.length === pData.size) {
                pData.value.set(fromCache);
                return VFS.SQLITE_OK;
            }
            try {
                // Read as many blocks as necessary to satisfy the read request.
                // Usually a read fits within a single write but there is at least
                // one case - rollback after journal spill - where reads cross
                // write boundaries so we have to allow for that.
                const result = await db.run("readonly", async ({ blocks }) => {
                    let pDataOffset = 0;
                    while (pDataOffset < pData.value.length) {
                        // console.log("read", pDataOffset);
                        // Fetch the IndexedDB block for this file location.
                        const fileOffset = iOffset + pDataOffset;
                        if (!file.block0)
                            throw new Error("Block0 not present");
                        const block0 = file.block0;
                        const block = await (async () => {
                            if (fileOffset < block0.data.length) {
                                return file.block0;
                            }
                            else {
                                const res = await blocks.get(this.#bound(-fileOffset));
                                return "data" in res
                                    ? res
                                    : {
                                        data: res,
                                        offset: -fileOffset,
                                    };
                            }
                        })();
                        if (!block || block.data.length - block.offset <= fileOffset) {
                            pData.value.fill(0, pDataOffset);
                            return VFS.SQLITE_IOERR_SHORT_READ;
                        }
                        const buffer = pData.value.subarray(pDataOffset);
                        const blockOffset = fileOffset + block.offset;
                        const nBytesToCopy = Math.min(Math.max(block.data.length - blockOffset, 0), // source bytes
                        buffer.length); // destination bytes
                        buffer.set(block.data.subarray(blockOffset, blockOffset + nBytesToCopy));
                        pDataOffset += nBytesToCopy;
                    }
                    return VFS.SQLITE_OK;
                });
                return result;
            }
            catch (e) {
                console.error(e);
                return VFS.SQLITE_IOERR;
            }
        });
    }
    xWrite(fileId, pData, iOffset) {
        const { file } = this.getFileStateByIdOrThrow(fileId);
        log$2(`xWrite ${file.name} ${pData.value.length} ${iOffset}`);
        try {
            if (!file.block0) {
                throw new Error("block0 not present");
            }
            // Convert the write directly into an IndexedDB object. Our assumption
            // is that SQLite will only overwrite data with an xWrite of the same
            // offset and size unless the database page size changes, except when
            // changing database page size which is handled by #reblockIfNeeded().
            file.block0.fileSize = Math.max(file.block0.fileSize, iOffset + pData.value.length);
            const block = (() => {
                if (iOffset === 0) {
                    file.block0.data = pData.value.slice();
                    return file.block0;
                }
                return {
                    offset: -iOffset,
                    data: pData.value.slice(),
                };
            })();
            this.blockToWrite.set(iOffset !== 0 ? -iOffset : 0, block);
            return VFS.SQLITE_OK;
        }
        catch (e) {
            console.error(e);
            return VFS.SQLITE_IOERR;
        }
    }
    xTruncate(fileId, iSize) {
        const { file, db } = this.getFileStateByIdOrThrow(fileId);
        if (!file)
            throw new Error("File not found");
        log$2(`xTruncate ${file.name} ${iSize}`);
        try {
            if (!file.block0) {
                throw new Error("block0 not present");
            }
            Object.assign(file.block0, {
                fileSize: iSize,
                data: file.block0.data.slice(0, iSize),
            });
            // Delete all blocks beyond the file size and update metadata.
            // This is never called within a transaction.
            const block0 = { ...file.block0 };
            void db.run("readwrite", async ({ blocks }) => {
                await blocks.delete(this.#bound(-Infinity, -iSize));
                await blocks.put(block0, 0);
            });
            return VFS.SQLITE_OK;
        }
        catch (e) {
            console.error(e);
            return VFS.SQLITE_IOERR;
        }
    }
    xSync(fileId, flags) {
        const { file, db } = this.getFileStateByIdOrThrow(fileId);
        if (!file)
            throw new Error("File not found");
        log$2(`xSync ${file.name} ${flags}`);
        try {
            if (this.options.durability !== "relaxed") {
                return this.handleAsync(async () => {
                    console.log("sync!");
                    await db.sync();
                    return VFS.SQLITE_OK;
                });
            }
            return VFS.SQLITE_OK;
        }
        catch (e) {
            console.error(e);
            return VFS.SQLITE_IOERR;
        }
    }
    xFileSize(fileId, pSize64) {
        const { file } = this.getFileStateByIdOrThrow(fileId);
        log$2(`xFileSize ${file.name}`);
        if (!file.block0) {
            throw new Error("block0 not present");
        }
        pSize64.set(file.block0.fileSize);
        return VFS.SQLITE_OK;
    }
    xLock(fileId, flags) {
        return this.handleAsync(async () => {
            const { file, db } = this.getFileStateByIdOrThrow(fileId);
            if (!file)
                throw new Error("File not found");
            log$2(`xLock ${file.name} ${flags}`);
            try {
                // Acquire the lock.
                log$2("getting lock...");
                const result = await this.webLocks.lock(file.name, flags);
                log$2(result);
                if (result === VFS.SQLITE_OK && flags === VFS.SQLITE_LOCK_SHARED) {
                    // Update block 0 in case another connection changed it.
                    file.block0 = await db.run("readonly", ({ blocks }) => blocks.get(this.#bound(0)));
                }
                return result;
            }
            catch (e) {
                console.error(e);
                return VFS.SQLITE_IOERR;
            }
        });
    }
    xUnlock(fileId, flags) {
        return this.handleAsync(async () => {
            const { file } = this.getFileStateByIdOrThrow(fileId);
            log$2(`xUnlock ${file.name} ${flags}`);
            try {
                return this.webLocks.unlock(file.name, flags);
            }
            catch (e) {
                console.error(e);
                return VFS.SQLITE_IOERR;
            }
        });
    }
    xSectorSize(fileId) {
        return SECTOR_SIZE$1;
    }
    xDeviceCharacteristics(fileId) {
        return (VFS.SQLITE_IOCAP_BATCH_ATOMIC |
            VFS.SQLITE_IOCAP_SAFE_APPEND |
            VFS.SQLITE_IOCAP_SEQUENTIAL |
            VFS.SQLITE_IOCAP_UNDELETABLE_WHEN_OPEN);
    }
    xFileControl(fileId, op) {
        const { file, db } = this.getFileStateByIdOrThrow(fileId);
        log$2(`xFileControl ${file.name} ${op}`);
        switch (op) {
            case 11: // SQLITE_FCNTL_OVERWRITE
                // This called on VACUUM. Set a flag so we know whether to check
                // later if the page size changed.
                file.overwrite = true;
                return VFS.SQLITE_OK;
            case 21: {
                // SQLITE_FCNTL_SYNC
                console.log("writing blocks...", Array.from(this.blockToWrite.keys()).length);
                const toWriteBlocks = this.blockToWrite;
                this.blockToWrite = new Map();
                return this.handleAsync(async () => {
                    console.time("blockToWrite");
                    const idb = await db.dbReady;
                    await new Promise((resolve, reject) => {
                        // @ts-expect-error lib dom misses third argument
                        const tx = idb.transaction("blocks", "readwrite", {
                            durability: "relaxed",
                        });
                        const blocksStore = tx.objectStore("blocks");
                        for (const k of toWriteBlocks.keys()) {
                            const toWrite = toWriteBlocks.get(k);
                            blocksStore.put(toWrite.offset === 0 ? toWrite : toWrite.data, Math.ceil(toWrite.offset / blockSize));
                        }
                        tx.addEventListener("complete", () => {
                            resolve();
                        });
                        tx.addEventListener("abort", reject);
                        tx.addEventListener("error", reject);
                    });
                    //           await db.run("readwrite", async ({ blocks }) => {
                    //             // assuming toWriteBlocks is a Map...
                    //             for (const k of toWriteBlocks.keys()) {
                    //               const toWrite = toWriteBlocks.get(k)!;
                    //               void blocks.put(
                    //                 toWrite.offset === 0 ? toWrite : toWrite.data,
                    //                 Math.ceil(toWrite.offset / blockSize)
                    //               );
                    //             }
                    //             await db.sync();
                    //           });
                    // await db.run("readwrite", async ({ blocks }) => {
                    //   await Promise.all(
                    //     Array.from(this.blockToWrite.keys()).map(async (k) => {
                    //       const toWrite = toWriteBlocks.get(k)!;
                    //       await blocks.put(toWrite);
                    //     })
                    //   );
                    // });
                    console.timeEnd("blockToWrite");
                    this.blockToWrite = new Map();
                    // This is called at the end of each database transaction, whether
                    // it is batch atomic or not. Handle page size changes here.
                    if (file.overwrite) {
                        // As an optimization we only check for and handle a page file
                        // changes if we know a VACUUM has been done because handleAsync()
                        // has to unwind and rewind the stack. We must be sure to follow
                        // the same conditional path in both calls.
                        try {
                            return this.handleAsync(async () => {
                                await this.#reblockIfNeeded(file, db);
                                return VFS.SQLITE_OK;
                            });
                        }
                        catch (e) {
                            console.error(e);
                            return VFS.SQLITE_IOERR;
                        }
                    }
                    return VFS.SQLITE_OK;
                });
            }
            case 22: // SQLITE_FCNTL_COMMIT_PHASETWO
                // This is called after a commit is completed.
                file.overwrite = false;
                return VFS.SQLITE_OK;
            case 31: // SQLITE_FCNTL_BEGIN_ATOMIC_WRITE
                return VFS.SQLITE_OK;
            case 32: // SQLITE_FCNTL_COMMIT_ATOMIC_WRITE
                console.log("commit!!");
                try {
                    if (!file.block0)
                        throw new Error("block0 not present!");
                    const block0 = { ...file.block0 };
                    block0.data = block0.data.slice();
                    void db.run("readwrite", async ({ blocks }) => {
                        // Write block 0 to commit the new version.
                        await blocks.put(block0, 0);
                    });
                    return VFS.SQLITE_OK;
                }
                catch (e) {
                    console.error(e);
                    return VFS.SQLITE_IOERR;
                }
            case 33: // SQLITE_FCNTL_ROLLBACK_ATOMIC_WRITE
                return this.handleAsync(async () => {
                    try {
                        this.blockToWrite = new Map();
                        return VFS.SQLITE_OK;
                    }
                    catch (e) {
                        console.error(e);
                        return VFS.SQLITE_IOERR;
                    }
                });
            default:
                return VFS.SQLITE_NOTFOUND;
        }
    }
    xAccess(name, flags, pResOut) {
        // const { db } = this.getFileStateByPathOrThrow(name);
        pResOut.set(0);
        return VFS.SQLITE_OK;
        // console.log("xAccess", name, { db });
        // return this.handleAsync(async () => {
        //   try {
        //     const path = new URL(name, "file://localhost/").pathname;
        //     log(`xAccess ${path} ${flags}`);
        //     // Check if block 0 exists.
        //     const key = await db.run("readonly", ({ blocks }) =>
        //       blocks.getKey(this.#bound({ path }, 0))
        //     );
        //     console.log(key ? 1 : 0);
        //     pResOut.set(key ? 1 : 0);
        //     return VFS.SQLITE_OK;
        //   } catch (e) {
        //     console.error(e);
        //     return VFS.SQLITE_IOERR;
        //   }
        // });
    }
    xDelete(name, syncDir) {
        const { db } = this.getFileStateByPathOrThrow(name);
        return this.handleAsync(async () => {
            const path = new URL(name, "file://localhost/").pathname;
            try {
                void db.run("readwrite", ({ blocks }) => blocks.delete(IDBKeyRange.bound([path], [path, []])));
                if (syncDir) {
                    await db.sync();
                }
                return VFS.SQLITE_OK;
            }
            catch (e) {
                console.error(e);
                return VFS.SQLITE_IOERR;
            }
        });
    }
    async purge(path) {
        // const start = Date.now();
        // await this.idb.run("readwrite", async ({ blocks }) => {
        //   const purgeBlock = await blocks.get([path, "purge", 0]);
        //   if (purgeBlock) {
        //     for (const [pageOffset, version] of purgeBlock.data) {
        //       blocks.delete(
        //         IDBKeyRange.bound(
        //           [path, pageOffset, version],
        //           [path, pageOffset, Infinity],
        //           true,
        //           false
        //         )
        //       );
        //     }
        //     await blocks.delete([path, "purge", 0]);
        //   }
        //   log(
        //     `purge ${path} ${purgeBlock?.data.size ?? 0} pages in ${
        //       Date.now() - start
        //     } ms`
        //   );
        // });
    }
    #maybePurge(path, nPages) {
        if (this.options.purge === "manual" ||
            this.pendingPurges.has(path) ||
            nPages < this.options.purgeAtLeast) {
            // No purge needed.
            return;
        }
        if (globalThis.requestIdleCallback) {
            globalThis.requestIdleCallback(() => {
                void this.purge(path);
                this.pendingPurges.delete(path);
            });
        }
        else {
            setTimeout(() => {
                void this.purge(path);
                this.pendingPurges.delete(path);
            });
        }
        this.pendingPurges.add(path);
    }
    #bound(begin, end = 0) {
        return IDBKeyRange.bound(Math.ceil(begin / blockSize), end);
    }
    // The database page size can be changed with PRAGMA page_size and VACUUM.
    // The updated file will be overwritten with a regular transaction using
    // the old page size. After that it will be read and written using the
    // new page size, so the IndexedDB objects must be combined or split
    // appropriately.
    async #reblockIfNeeded(file, db) {
        // if (!file.block0) throw new Error("Block0 not present");
        // const oldPageSize = file.block0.data.length;
        // if (oldPageSize < 18) return; // no page size defined
        // const view = new DataView(
        //   file.block0.data.buffer,
        //   file.block0.data.byteOffset
        // );
        // let newPageSize = view.getUint16(16);
        // if (newPageSize === 1) newPageSize = 65536;
        // if (newPageSize === oldPageSize) return; // no page size change
        // const maxPageSize = Math.max(oldPageSize, newPageSize);
        // const nOldPages = maxPageSize / oldPageSize;
        // const nNewPages = maxPageSize / newPageSize;
        // const newPageCount = view.getUint32(28);
        // const fileSize = newPageCount * newPageSize;
        // await db.run("readwrite", async ({ blocks }) => {
        //   // When the block size changes, the entire file is rewritten. Delete
        //   // all blocks older than block 0 to leave a single version at every
        //   // offset.
        //   const keys = await blocks
        //     .index("version")
        //     .getAllKeys(
        //       IDBKeyRange.bound([file.path, version + 1], [file.path, Infinity])
        //     );
        //   for (const key of keys) {
        //     void blocks.delete(key);
        //   }
        //   await blocks.delete([file.path, "purge", 0]);
        //   // Do the conversion in chunks of the larger of the page sizes.
        //   for (let iOffset = 0; iOffset < fileSize; iOffset += maxPageSize) {
        //     // Fetch nOldPages. They can be fetched in one request because
        //     // there is now a single version in the file.
        //     const oldPages = await blocks.getAll(
        //       IDBKeyRange.lowerBound([
        //         file.path,
        //         -(iOffset + maxPageSize),
        //         Infinity,
        //       ]),
        //       nOldPages
        //     );
        //     for (const oldPage of oldPages) {
        //       void blocks.delete([oldPage.path, oldPage.offset, oldPage.version]);
        //     }
        //     // Convert to new pages.
        //     if (nNewPages === 1) {
        //       // Combine nOldPages old pages into a new page.
        //       const buffer = new Int8Array(newPageSize);
        //       for (const oldPage of oldPages) {
        //         buffer.set(oldPage.data, -(iOffset + oldPage.offset));
        //       }
        //       const newPage = {
        //         path: file.path,
        //         offset: -iOffset,
        //         version,
        //         data: buffer,
        //       };
        //       if (newPage.offset === 0) {
        //         file.block0 = { ...newPage, fileSize };
        //       }
        //       await blocks.put(newPage);
        //     } else {
        //       // Split an old page into nNewPages new pages.
        //       const oldPage = oldPages[0];
        //       for (let i = 0; i < nNewPages; ++i) {
        //         const offset = -(iOffset + i * newPageSize);
        //         if (-offset >= fileSize) break;
        //         const newPage = {
        //           path: oldPage.path,
        //           offset,
        //           version,
        //           data: oldPage.data.subarray(
        //             i * newPageSize,
        //             (i + 1) * newPageSize
        //           ),
        //         };
        //         if (newPage.offset === 0) {
        //           file.block0 = { ...newPage, fileSize };
        //         }
        //         await blocks.put(newPage);
        //       }
        //     }
        //   }
        // });
    }
    getFileStateByPathOrThrow(path) {
        const dbName = path.replace(/-journal$/, "").replace(/-wal$/, "");
        const fileId = this.dbNameFileIdMap.get(dbName);
        if (!fileId) {
            throw new Error(`Failed to find db state for path: ${path}`);
        }
        return this.getFileStateByIdOrThrow(fileId);
    }
    getFileStateByIdOrThrow(fileId) {
        const state = this.filesState.get(fileId);
        if (!state)
            throw new Error("File not found");
        return state;
    }
}
function openDatabase$2(idbDatabaseName) {
    return new Promise((resolve, reject) => {
        const request = globalThis.indexedDB.open(idbDatabaseName, 5);
        request.addEventListener("upgradeneeded", () => {
            request.result.createObjectStore("blocks");
        });
        request.addEventListener("success", () => {
            resolve(request.result);
        });
        request.addEventListener("error", () => {
            reject(request.error);
        });
    });
}

const SECTOR_SIZE = 512;
const DEFAULT_OPTIONS$1 = {
    durability: "default",
    purge: "deferred",
    purgeAtLeast: 16,
};
function log$1(...args) {
    // console.debug(...args);
}
// This sample VFS stores optionally versioned writes to IndexedDB, which
// it uses with the SQLite xFileControl() batch atomic write feature.
class IDBBatchAtomicVFS extends VFS.Base {
    name;
    options;
    mapIdToFile = new Map();
    idb;
    webLocks = new WebLocks();
    pendingPurges = new Set();
    blockToWrite = new Map();
    constructor(idbDatabaseName = "wa-sqlite", options = DEFAULT_OPTIONS$1) {
        super();
        this.name = idbDatabaseName;
        this.options = Object.assign({}, DEFAULT_OPTIONS$1, options);
        this.idb = new IDBContext(openDatabase$1(idbDatabaseName), {
            durability: this.options.durability,
        });
    }
    xOpen(name, fileId, flags, pOutFlags) {
        return this.handleAsync(async () => {
            if (name === null)
                name = `null_${fileId}`;
            log$1(`xOpen ${name} 0x${fileId.toString(16)} 0x${flags.toString(16)}`);
            try {
                // Filenames can be URLs, possibly with query parameters.
                const url = new URL(name, "http://localhost/");
                const file = {
                    path: url.pathname,
                    flags,
                    block0: undefined,
                };
                this.mapIdToFile.set(fileId, file);
                // Read the first block, which also contains the file metadata.
                file.block0 = await this.idb.run("readonly", ({ blocks }) => {
                    return blocks.get(this.#bound(file, 0));
                });
                if (!file.block0) {
                    // File doesn't exist, create if requested.
                    if (flags & VFS.SQLITE_OPEN_CREATE) {
                        file.block0 = {
                            path: file.path,
                            offset: 0,
                            version: 0,
                            data: new Int8Array(0),
                            fileSize: 0,
                        };
                        // Write metadata block to IndexedDB.
                        void this.idb.run("readwrite", ({ blocks }) => blocks.put(file.block0));
                        await this.idb.sync();
                    }
                    else {
                        throw new Error(`file not found: ${file.path}`);
                    }
                }
                pOutFlags.set(flags & VFS.SQLITE_OPEN_READONLY);
                return VFS.SQLITE_OK;
            }
            catch (e) {
                console.error(e);
                return VFS.SQLITE_CANTOPEN;
            }
        });
    }
    xClose(fileId) {
        return this.handleAsync(async () => {
            try {
                const file = this.mapIdToFile.get(fileId);
                if (file) {
                    log$1(`xClose ${file.path}`);
                    this.mapIdToFile.delete(fileId);
                    if (file.flags & VFS.SQLITE_OPEN_DELETEONCLOSE) {
                        void this.idb.run("readwrite", async ({ blocks }) => {
                            await blocks.delete(IDBKeyRange.bound([file.path], [file.path, []]));
                        });
                    }
                }
                return VFS.SQLITE_OK;
            }
            catch (e) {
                console.error(e);
                return VFS.SQLITE_IOERR;
            }
        });
    }
    xRead(fileId, pData, iOffset) {
        return this.handleAsync(async () => {
            const file = this.mapIdToFile.get(fileId);
            if (!file)
                throw new Error("File not found");
            log$1(`xRead ${file.path} ${pData.value.length} ${iOffset}`);
            const fromCache = this.blockToWrite.get(-iOffset)?.data;
            if (fromCache && fromCache.length === pData.size) {
                pData.value.set(fromCache);
                return VFS.SQLITE_OK;
            }
            try {
                // Read as many blocks as necessary to satisfy the read request.
                // Usually a read fits within a single write but there is at least
                // one case - rollback after journal spill - where reads cross
                // write boundaries so we have to allow for that.
                const result = await this.idb.run("readonly", async ({ blocks }) => {
                    let pDataOffset = 0;
                    while (pDataOffset < pData.value.length) {
                        // Fetch the IndexedDB block for this file location.
                        const fileOffset = iOffset + pDataOffset;
                        if (!file.block0)
                            throw new Error("Block0 not present");
                        const block = fileOffset < file.block0.data.length
                            ? file.block0
                            : await blocks.get(this.#bound(file, -fileOffset));
                        if (!block || block.data.length - block.offset <= fileOffset) {
                            pData.value.fill(0, pDataOffset);
                            return VFS.SQLITE_IOERR_SHORT_READ;
                        }
                        const buffer = pData.value.subarray(pDataOffset);
                        const blockOffset = fileOffset + block.offset;
                        const nBytesToCopy = Math.min(Math.max(block.data.length - blockOffset, 0), // source bytes
                        buffer.length); // destination bytes
                        buffer.set(block.data.subarray(blockOffset, blockOffset + nBytesToCopy));
                        pDataOffset += nBytesToCopy;
                    }
                    return VFS.SQLITE_OK;
                });
                return result;
            }
            catch (e) {
                console.error(e);
                return VFS.SQLITE_IOERR;
            }
        });
    }
    xWrite(fileId, pData, iOffset) {
        const file = this.mapIdToFile.get(fileId);
        if (!file)
            throw new Error("File not found");
        log$1(`xWrite ${file.path} ${pData.value.length} ${iOffset}`);
        try {
            if (!file.block0) {
                throw new Error("block0 not present");
            }
            // Convert the write directly into an IndexedDB object. Our assumption
            // is that SQLite will only overwrite data with an xWrite of the same
            // offset and size unless the database page size changes, except when
            // changing database page size which is handled by #reblockIfNeeded().
            const prevFileSize = file.block0.fileSize;
            file.block0.fileSize = Math.max(file.block0.fileSize, iOffset + pData.value.length);
            const block = (() => {
                if (iOffset === 0) {
                    file.block0.data = pData.value.slice();
                    return file.block0;
                }
                else {
                    return {
                        path: file.path,
                        offset: -iOffset,
                        version: file.block0.version,
                        data: pData.value.slice(),
                    };
                }
            })();
            if (file.changedPages) {
                // This write is part of a batch atomic write. All writes in the
                // batch have a new version, so update the changed list to allow
                // old versions to be eventually deleted.
                if (prevFileSize === file.block0.fileSize) {
                    file.changedPages.add(-iOffset);
                }
                // Defer writing block 0 to IndexedDB until batch commit.
                if (iOffset !== 0) {
                    // this.idb.run("readwrite", ({ blocks }) => blocks.put(block));
                }
            }
            else {
                // Not a batch atomic write so write through.
                // this.idb.run("readwrite", ({ blocks }) => blocks.put(block));
            }
            this.blockToWrite.set(iOffset !== 0 ? -iOffset : 0, block);
            return VFS.SQLITE_OK;
        }
        catch (e) {
            console.error(e);
            return VFS.SQLITE_IOERR;
        }
    }
    xTruncate(fileId, iSize) {
        const file = this.mapIdToFile.get(fileId);
        if (!file)
            throw new Error("File not found");
        log$1(`xTruncate ${file.path} ${iSize}`);
        try {
            if (!file.block0) {
                throw new Error("block0 not present");
            }
            Object.assign(file.block0, {
                fileSize: iSize,
                data: file.block0.data.slice(0, iSize),
            });
            // Delete all blocks beyond the file size and update metadata.
            // This is never called within a transaction.
            const block0 = Object.assign({}, file.block0);
            void this.idb.run("readwrite", async ({ blocks }) => {
                await blocks.delete(this.#bound(file, -Infinity, -iSize));
                await blocks.put(block0);
            });
            return VFS.SQLITE_OK;
        }
        catch (e) {
            console.error(e);
            return VFS.SQLITE_IOERR;
        }
    }
    xSync(fileId, flags) {
        const file = this.mapIdToFile.get(fileId);
        if (!file)
            throw new Error("File not found");
        log$1(`xSync ${file.path} ${flags}`);
        try {
            if (this.options.durability !== "relaxed") {
                return this.handleAsync(async () => {
                    await this.idb.sync();
                    return VFS.SQLITE_OK;
                });
            }
            return VFS.SQLITE_OK;
        }
        catch (e) {
            console.error(e);
            return VFS.SQLITE_IOERR;
        }
    }
    xFileSize(fileId, pSize64) {
        const file = this.mapIdToFile.get(fileId);
        if (!file)
            throw new Error("File not found");
        log$1(`xFileSize ${file.path}`);
        if (!file.block0) {
            throw new Error("block0 not present");
        }
        pSize64.set(file.block0.fileSize);
        return VFS.SQLITE_OK;
    }
    xLock(fileId, flags) {
        return this.handleAsync(async () => {
            const file = this.mapIdToFile.get(fileId);
            if (!file)
                throw new Error("File not found");
            log$1(`xLock ${file.path} ${flags}`);
            try {
                // Acquire the lock.
                log$1("getting lock...");
                const result = await this.webLocks.lock(file.path, flags);
                log$1(result);
                if (result === VFS.SQLITE_OK && flags === VFS.SQLITE_LOCK_SHARED) {
                    // Update block 0 in case another connection changed it.
                    file.block0 = await this.idb.run("readonly", ({ blocks }) => {
                        return blocks.get(this.#bound(file, 0));
                    });
                }
                return result;
            }
            catch (e) {
                console.error(e);
                return VFS.SQLITE_IOERR;
            }
        });
    }
    xUnlock(fileId, flags) {
        return this.handleAsync(async () => {
            const file = this.mapIdToFile.get(fileId);
            if (!file)
                throw new Error("File not found");
            log$1(`xUnlock ${file.path} ${flags}`);
            try {
                return this.webLocks.unlock(file.path, flags);
            }
            catch (e) {
                console.error(e);
                return VFS.SQLITE_IOERR;
            }
        });
    }
    xSectorSize(fileId) {
        return SECTOR_SIZE;
    }
    xDeviceCharacteristics(fileId) {
        return (VFS.SQLITE_IOCAP_BATCH_ATOMIC |
            VFS.SQLITE_IOCAP_SAFE_APPEND |
            VFS.SQLITE_IOCAP_SEQUENTIAL |
            VFS.SQLITE_IOCAP_UNDELETABLE_WHEN_OPEN);
    }
    xFileControl(fileId, op) {
        const file = this.mapIdToFile.get(fileId);
        if (!file)
            throw new Error("File not found");
        log$1(`xFileControl ${file.path} ${op}`);
        switch (op) {
            case 11: //SQLITE_FCNTL_OVERWRITE
                // This called on VACUUM. Set a flag so we know whether to check
                // later if the page size changed.
                file.overwrite = true;
                return VFS.SQLITE_OK;
            case 21: {
                // SQLITE_FCNTL_SYNC
                console.log("writing blocks...", Array.from(this.blockToWrite.keys()).length);
                return this.handleAsync(async () => {
                    console.time("blockToWrite");
                    await Promise.all(Array.from(this.blockToWrite.keys()).map(async (k) => {
                        await this.idb.run("readwrite", async ({ blocks }) => {
                            // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
                            const block = this.blockToWrite.get(k);
                            await blocks.put(block);
                        });
                    }));
                    console.timeEnd("blockToWrite");
                    this.blockToWrite = new Map();
                    // This is called at the end of each database transaction, whether
                    // it is batch atomic or not. Handle page size changes here.
                    if (file.overwrite) {
                        // As an optimization we only check for and handle a page file
                        // changes if we know a VACUUM has been done because handleAsync()
                        // has to unwind and rewind the stack. We must be sure to follow
                        // the same conditional path in both calls.
                        try {
                            return this.handleAsync(async () => {
                                await this.#reblockIfNeeded(file);
                                return VFS.SQLITE_OK;
                            });
                        }
                        catch (e) {
                            console.error(e);
                            return VFS.SQLITE_IOERR;
                        }
                    }
                    return VFS.SQLITE_OK;
                });
            }
            case 22: // SQLITE_FCNTL_COMMIT_PHASETWO
                // This is called after a commit is completed.
                file.overwrite = false;
                return VFS.SQLITE_OK;
            case 31: // SQLITE_FCNTL_BEGIN_ATOMIC_WRITE
                return this.handleAsync(async () => {
                    try {
                        if (!file.block0) {
                            throw new Error("block0 not present");
                        }
                        // Prepare a new version for IndexedDB blocks.
                        file.block0.version--;
                        file.changedPages = new Set();
                        // Clear blocks from abandoned transactions that would conflict
                        // with the new transaction.
                        await this.idb.run("readwrite", async ({ blocks }) => {
                            if (!file.block0) {
                                throw new Error("block0 not present");
                            }
                            const keys = await blocks
                                .index("version")
                                .getAllKeys(IDBKeyRange.bound([file.path], [file.path, file.block0.version]));
                            for (const key of keys) {
                                void blocks.delete(key);
                            }
                        });
                        return VFS.SQLITE_OK;
                    }
                    catch (e) {
                        console.error(e);
                        return VFS.SQLITE_IOERR;
                    }
                });
            case 32: // SQLITE_FCNTL_COMMIT_ATOMIC_WRITE
                try {
                    const block0 = Object.assign({}, file.block0);
                    block0.data = block0.data.slice();
                    if (!file.changedPages) {
                        throw new Error("Files doesn't have changed pages");
                    }
                    const changedPages = file.changedPages;
                    file.changedPages = undefined;
                    void this.idb.run("readwrite", async ({ blocks }) => {
                        // Write block 0 to commit the new version.
                        void blocks.put(block0);
                        // Blocks to purge are saved in a special IndexedDB object with
                        // an "index" of "purge". Add pages changed by this transaction.
                        const purgeBlock = (await blocks.get([file.path, "purge", 0])) ?? {
                            path: file.path,
                            offset: "purge",
                            version: 0,
                            data: new Map(),
                            count: 0,
                        };
                        purgeBlock.count += changedPages.size;
                        for (const pageIndex of changedPages) {
                            purgeBlock.data.set(pageIndex, block0.version);
                        }
                        void blocks.put(purgeBlock);
                        this.#maybePurge(file.path, purgeBlock.count);
                    });
                    return VFS.SQLITE_OK;
                }
                catch (e) {
                    console.error(e);
                    return VFS.SQLITE_IOERR;
                }
            case 33: // SQLITE_FCNTL_ROLLBACK_ATOMIC_WRITE
                return this.handleAsync(async () => {
                    try {
                        // Restore original state. Objects for the abandoned version will
                        // be left in IndexedDB to be removed by the next atomic write
                        // transaction.
                        file.changedPages = undefined;
                        file.block0 = await this.idb.run("readonly", ({ blocks }) => {
                            if (!file.block0) {
                                throw new Error("block0 not present");
                            }
                            return blocks.get([file.path, 0, file.block0.version + 1]);
                        });
                        return VFS.SQLITE_OK;
                    }
                    catch (e) {
                        console.error(e);
                        return VFS.SQLITE_IOERR;
                    }
                });
            default:
                return VFS.SQLITE_NOTFOUND;
        }
    }
    xAccess(name, flags, pResOut) {
        return this.handleAsync(async () => {
            try {
                const path = new URL(name, "file://localhost/").pathname;
                log$1(`xAccess ${path} ${flags}`);
                // Check if block 0 exists.
                const key = await this.idb.run("readonly", ({ blocks }) => {
                    return blocks.getKey(this.#bound({ path }, 0));
                });
                pResOut.set(key ? 1 : 0);
                return VFS.SQLITE_OK;
            }
            catch (e) {
                console.error(e);
                return VFS.SQLITE_IOERR;
            }
        });
    }
    xDelete(name, syncDir) {
        return this.handleAsync(async () => {
            const path = new URL(name, "file://localhost/").pathname;
            try {
                void this.idb.run("readwrite", ({ blocks }) => {
                    return blocks.delete(IDBKeyRange.bound([path], [path, []]));
                });
                if (syncDir) {
                    await this.idb.sync();
                }
                return VFS.SQLITE_OK;
            }
            catch (e) {
                console.error(e);
                return VFS.SQLITE_IOERR;
            }
        });
    }
    async purge(path) {
        const start = Date.now();
        await this.idb.run("readwrite", async ({ blocks }) => {
            const purgeBlock = await blocks.get([path, "purge", 0]);
            if (purgeBlock) {
                for (const [pageOffset, version] of purgeBlock.data) {
                    void blocks.delete(IDBKeyRange.bound([path, pageOffset, version], [path, pageOffset, Infinity], true, false));
                }
                await blocks.delete([path, "purge", 0]);
            }
            log$1(`purge ${path} ${purgeBlock?.data.size ?? 0} pages in ${Date.now() - start} ms`);
        });
    }
    #maybePurge(path, nPages) {
        if (this.options.purge === "manual" ||
            this.pendingPurges.has(path) ||
            nPages < this.options.purgeAtLeast) {
            // No purge needed.
            return;
        }
        if (globalThis.requestIdleCallback) {
            globalThis.requestIdleCallback(() => {
                void this.purge(path);
                this.pendingPurges.delete(path);
            });
        }
        else {
            setTimeout(() => {
                void this.purge(path);
                this.pendingPurges.delete(path);
            });
        }
        this.pendingPurges.add(path);
    }
    #bound(file, begin, end = 0) {
        // Fetch newest block 0. For other blocks, use block 0 version.
        const version = (() => {
            if (!("block0" in file))
                return -Infinity;
            if (!begin || (file.block0 && -begin < file.block0.data.length)) {
                return -Infinity;
            }
            return file.block0?.version;
        })();
        return IDBKeyRange.bound([file.path, begin, version], [file.path, end, Infinity]);
    }
    // The database page size can be changed with PRAGMA page_size and VACUUM.
    // The updated file will be overwritten with a regular transaction using
    // the old page size. After that it will be read and written using the
    // new page size, so the IndexedDB objects must be combined or split
    // appropriately.
    async #reblockIfNeeded(file) {
        if (!file.block0)
            throw new Error("Block0 not present");
        const oldPageSize = file.block0.data.length;
        if (oldPageSize < 18)
            return; // no page size defined
        const view = new DataView(file.block0.data.buffer, file.block0.data.byteOffset);
        let newPageSize = view.getUint16(16);
        if (newPageSize === 1)
            newPageSize = 65536;
        if (newPageSize === oldPageSize)
            return; // no page size change
        const maxPageSize = Math.max(oldPageSize, newPageSize);
        const nOldPages = maxPageSize / oldPageSize;
        const nNewPages = maxPageSize / newPageSize;
        const newPageCount = view.getUint32(28);
        const fileSize = newPageCount * newPageSize;
        const version = file.block0.version;
        await this.idb.run("readwrite", async ({ blocks }) => {
            // When the block size changes, the entire file is rewritten. Delete
            // all blocks older than block 0 to leave a single version at every
            // offset.
            const keys = await blocks
                .index("version")
                .getAllKeys(IDBKeyRange.bound([file.path, version + 1], [file.path, Infinity]));
            for (const key of keys) {
                void blocks.delete(key);
            }
            void blocks.delete([file.path, "purge", 0]);
            // Do the conversion in chunks of the larger of the page sizes.
            for (let iOffset = 0; iOffset < fileSize; iOffset += maxPageSize) {
                // Fetch nOldPages. They can be fetched in one request because
                // there is now a single version in the file.
                const oldPages = await blocks.getAll(IDBKeyRange.lowerBound([
                    file.path,
                    -(iOffset + maxPageSize),
                    Infinity,
                ]), nOldPages);
                for (const oldPage of oldPages) {
                    void blocks.delete([oldPage.path, oldPage.offset, oldPage.version]);
                }
                // Convert to new pages.
                if (nNewPages === 1) {
                    // Combine nOldPages old pages into a new page.
                    const buffer = new Int8Array(newPageSize);
                    for (const oldPage of oldPages) {
                        buffer.set(oldPage.data, -(iOffset + oldPage.offset));
                    }
                    const newPage = {
                        path: file.path,
                        offset: -iOffset,
                        version,
                        data: buffer,
                    };
                    if (newPage.offset === 0) {
                        file.block0 = { ...newPage, fileSize };
                    }
                    void blocks.put(newPage);
                }
                else {
                    // Split an old page into nNewPages new pages.
                    const oldPage = oldPages[0];
                    for (let i = 0; i < nNewPages; ++i) {
                        const offset = -(iOffset + i * newPageSize);
                        if (-offset >= fileSize)
                            break;
                        const newPage = {
                            path: oldPage.path,
                            offset,
                            version,
                            data: oldPage.data.subarray(i * newPageSize, (i + 1) * newPageSize),
                        };
                        if (newPage.offset === 0) {
                            file.block0 = { ...newPage, fileSize };
                        }
                        void blocks.put(newPage);
                    }
                }
            }
        });
    }
}
function openDatabase$1(idbDatabaseName) {
    return new Promise((resolve, reject) => {
        const request = globalThis.indexedDB.open(idbDatabaseName, 5);
        request.addEventListener("upgradeneeded", function () {
            const blocks = request.result.createObjectStore("blocks", {
                keyPath: ["path", "offset", "version"],
            });
            blocks.createIndex("version", ["path", "version"]);
        });
        request.addEventListener("success", () => {
            resolve(request.result);
        });
        request.addEventListener("error", () => {
            reject(request.error);
        });
    });
}

function log(...args) {
    // console.debug(...args);
}
const DEFAULT_OPTIONS = { durability: "default" };
class IDCachedWritesVFS extends VFS.Base {
    name;
    filesState = new Map();
    dbNameFileIdMap = new Map();
    webLocks = new WebLocks();
    options;
    pendingWrites = new Map();
    cursors = new Map();
    constructor(name, options = DEFAULT_OPTIONS) {
        super();
        this.name = name;
        this.options = options;
    }
    xOpen(name, fileId, flags, pOutFlags) {
        return this.handleAsync(async () => {
            if (name === null)
                name = `null_${fileId}`;
            this.dbNameFileIdMap.set(name, fileId);
            const db = new IDBContext(openDatabase(name), {
                durability: this.options.durability,
            });
            log(`xOpen ${name} ${fileId} 0x${flags.toString(16)}`);
            try {
                // Filenames can be URLs, possibly with query parameters.
                const url = new URL(name, "file://localhost/");
                const file = {
                    path: url.pathname,
                    flags,
                    block0: await (async () => {
                        const res = await db.run("readonly", ({ blocks }) => blocks.get(this.bound(0)));
                        if (!res) {
                            // File doesn't exist, create if requested.
                            if (flags & VFS.SQLITE_OPEN_CREATE) {
                                const block0 = {
                                    offset: 0,
                                    data: new Int8Array(0),
                                    fileSize: 0,
                                };
                                // Write metadata block to IndexedDB.
                                await db.run("readwrite", ({ blocks }) => blocks.put(block0, 0));
                                await db.sync();
                                return block0;
                            }
                            else {
                                throw new Error(`file not found: ${name}`);
                            }
                        }
                        else {
                            return res;
                        }
                    })(),
                };
                this.filesState.set(fileId, {
                    file,
                    db,
                });
                pOutFlags.set(flags & VFS.SQLITE_OPEN_READONLY);
                return VFS.SQLITE_OK;
            }
            catch (e) {
                console.error(e);
                return VFS.SQLITE_CANTOPEN;
            }
        });
    }
    xClose(fileId) {
        return this.handleAsync(async () => {
            try {
                const state = this.filesState.get(fileId);
                if (state) {
                    const { file, db } = state;
                    log(`xClose ${file.path}`);
                    this.filesState.delete(fileId);
                    if (file.flags & VFS.SQLITE_OPEN_DELETEONCLOSE) {
                        await db.run("readwrite", async ({ blocks }) => {
                            await blocks.delete(this.bound(-Infinity));
                        });
                    }
                }
                return VFS.SQLITE_OK;
            }
            catch (e) {
                console.error(e);
                return VFS.SQLITE_IOERR;
            }
        });
    }
    cursorPromises = new Map();
    xRead(fileId, pData, iOffset) {
        return this.handleAsync(async () => {
            // TODO: not sure why, but if is not in handleAsync then code called twice.
            // Is it stack unwind/rewind?
            const { db, file } = this.getFileStateByIdOrThrow(fileId);
            log(`xRead ${file.path} ${pData.value.length} ${iOffset}`);
            const filePendingWrites = this.pendingWrites.get(fileId);
            const block = filePendingWrites?.get(iOffset);
            const blockSize = file.block0.data.length;
            if (block) {
                pData.value.set(block.data);
                return VFS.SQLITE_OK;
            }
            const dir = this.getReadDirection(fileId, blockSize);
            const waitCursor = async () => {
                const block = await new Promise((resolve, reject) => {
                    if (this.cursorPromises.has(fileId)) {
                        throw new Error("waitCursor() called but something else is already waiting");
                    }
                    this.cursorPromises.set(fileId, { resolve, reject });
                });
                // console.log({ block });
                const blockOffset = iOffset + block.offset;
                const nBytesToCopy = Math.min(Math.max(block.data.length - blockOffset, 0), // source bytes
                pData.value.length); // destination bytes
                pData.value.set(block.data.subarray(blockOffset, blockOffset + nBytesToCopy));
                if (nBytesToCopy < pData.value.length) {
                    pData.value.fill(0, nBytesToCopy, pData.value.length);
                    return VFS.SQLITE_IOERR_SHORT_READ;
                }
                return VFS.SQLITE_OK;
            };
            const cursor = this.cursors.get(fileId);
            if (cursor) {
                const key = cursor.key;
                // console.log({ key });
                if ((cursor.direction === "next" &&
                    key < -iOffset &&
                    -iOffset < key + 100 * blockSize) ||
                    (cursor.direction === "prev" &&
                        key - 100 * blockSize < -iOffset &&
                        -iOffset < key)) {
                    cursor.advance(Math.abs(Math.ceil((-iOffset - key) / blockSize)));
                    return waitCursor();
                }
                else {
                    // console.log(
                    //   "cursor miss(",
                    //   cursor.direction,
                    //   cursor.direction === "next"
                    //     ? [key, -iOffset, key + 100 * blockSize]
                    //     : [key - 100 * blockSize, -iOffset, key]
                    // );
                    this.cursors.delete(fileId);
                }
            }
            else {
                if (dir) {
                    this.prevReads.delete(fileId);
                    const keyRange = (() => {
                        if (dir === "prev") {
                            return IDBKeyRange.upperBound(-iOffset);
                        }
                        else {
                            return IDBKeyRange.lowerBound(-iOffset);
                        }
                    })();
                    const idb = await db.dbReady;
                    // @ts-expect-error lib dom misses third argument
                    const tx = idb.transaction("blocks", "readwrite", {
                        durability: "relaxed",
                    });
                    const blocksStore = tx.objectStore("blocks");
                    const req = blocksStore.openCursor(keyRange, dir);
                    // console.warn("OPEN CURSOR", -iOffset, dir);
                    req.onsuccess = (e) => {
                        let cursor = e.target.result;
                        this.cursors.set(fileId, cursor);
                        const promise = this.cursorPromises.get(fileId);
                        if (!promise)
                            throw new Error("Got data from cursor but nothing is waiting it");
                        promise.resolve(cursor ? cursor.value : null);
                        this.cursorPromises.delete(fileId);
                    };
                    req.onerror = (e) => {
                        console.log("Cursor failure:", e);
                        const promise = this.cursorPromises.get(fileId);
                        if (!promise)
                            throw new Error("Got data from cursor but nothing is waiting it");
                        promise.reject(e);
                        this.cursorPromises.delete(fileId);
                    };
                    return await waitCursor();
                }
                else {
                    const res = this.prevReads.get(fileId);
                    if (!res) {
                        this.prevReads.set(fileId, [0, 0, -iOffset]);
                    }
                    else {
                        res.push(-iOffset);
                        res.shift();
                    }
                }
            }
            try {
                const block = await db.run("readonly", ({ blocks }) => {
                    return blocks.get(this.bound(-iOffset));
                });
                const blockOffset = iOffset + block.offset;
                const nBytesToCopy = Math.min(Math.max(block.data.length - blockOffset, 0), // source bytes
                pData.value.length); // destination bytes
                pData.value.set(block.data.subarray(blockOffset, blockOffset + nBytesToCopy));
                if (nBytesToCopy < pData.value.length) {
                    pData.value.fill(0, nBytesToCopy, pData.value.length);
                    return VFS.SQLITE_IOERR_SHORT_READ;
                }
                return VFS.SQLITE_OK;
            }
            catch (e) {
                console.error(e);
                return VFS.SQLITE_IOERR;
            }
        });
    }
    // The ideas is taken from absurd-sql. Very smart!
    // Also this might be helpful https://nolanlawson.com/2021/08/22/speeding-up-indexeddb-reads-and-writes/
    prevReads = new Map();
    getReadDirection(fileId, blockSize) {
        // There are a two ways we can read data: a direct `get` request
        // or opening a cursor and iterating through data. We don't know
        // what future reads look like, so we don't know the best strategy
        // to pick. Always choosing one strategy forgoes a lot of
        // optimization, because iterating with a cursor is a lot faster
        // than many `get` calls. On the other hand, opening a cursor is
        // slow, and so is calling `advance` to move a cursor over a huge
        // range (like moving it 1000 items later), so many `get` calls would
        // be faster. In general:
        //
        // * Many `get` calls are faster when doing random accesses
        // * Iterating with a cursor is faster if doing mostly sequential
        //   accesses
        //
        // We implement a heuristic and keeps track of the last 3 reads
        // and detects when they are mostly sequential. If they are, we
        // open a cursor and start reading by iterating it. If not, we do
        // direct `get` calls.
        //
        // On top of all of this, each browser has different perf
        // characteristics. We will probably want to make these thresholds
        // configurable so the user can change them per-browser if needed,
        // as well as fine-tuning them for their usage of sqlite.
        let prevReads = this.prevReads.get(fileId);
        if (prevReads) {
            // Has there been 3 forward sequential reads within 10 blocks?
            if (prevReads[0] < prevReads[1] &&
                prevReads[1] < prevReads[2] &&
                Math.abs(prevReads[2] - prevReads[0]) < 10 * blockSize) {
                return "next";
            }
            // Has there been 3 backwards sequential reads within 10 blocks?
            if (prevReads[0] > prevReads[1] &&
                prevReads[1] > prevReads[2] &&
                Math.abs(prevReads[0] - prevReads[2]) < 10 * blockSize) {
                return "prev";
            }
        }
        return null;
    }
    xWrite(fileId, pData, iOffset) {
        return this.handleAsync(async () => {
            const { file, db } = this.getFileStateByIdOrThrow(fileId);
            log(`xWrite ${file.path} ${pData.value.length} ${iOffset}`);
            try {
                // Convert the write directly into an IndexedDB object.
                const block = {
                    offset: -iOffset,
                    data: pData.value.slice(),
                };
                file.block0.fileSize = Math.max(file.block0.fileSize, iOffset + pData.value.length);
                if (iOffset === 0) {
                    file.block0.data = pData.value.slice();
                }
                const filePendingWrites = this.pendingWrites.get(fileId);
                if (filePendingWrites) {
                    if (iOffset === 0) {
                        filePendingWrites.set(0, file.block0);
                    }
                    else {
                        filePendingWrites.set(iOffset, block);
                    }
                    return VFS.SQLITE_OK;
                }
                // console.log("put", block, block.offset);
                await db.run("readwrite", async ({ blocks }) => {
                    if (iOffset === 0) {
                        await blocks.put(file.block0, 0);
                    }
                    else {
                        await blocks.put(block, block.offset);
                    }
                });
                return VFS.SQLITE_OK;
            }
            catch (e) {
                console.error(e);
                return VFS.SQLITE_IOERR;
            }
        });
    }
    xTruncate(fileId, iSize) {
        const { file, db } = this.getFileStateByIdOrThrow(fileId);
        log(`xTruncate ${file.path} ${iSize}`);
        try {
            const block = {
                offset: 0,
                data: new Int8Array(0),
                fileSize: 0,
            };
            file.block0 = block;
            void db.run("readwrite", async ({ blocks }) => {
                await blocks.delete(this.bound(-Infinity, -iSize));
                await blocks.put(block, 0);
            });
            return VFS.SQLITE_OK;
        }
        catch (e) {
            console.error(e);
            return VFS.SQLITE_IOERR;
        }
    }
    xSync(fileId, flags) {
        const { file, db } = this.getFileStateByIdOrThrow(fileId);
        if (this.options.durability !== "relaxed") {
            return this.handleAsync(async () => {
                log(`xSync ${file.path} ${flags}`);
                try {
                    await db.sync();
                    return VFS.SQLITE_OK;
                }
                catch (e) {
                    console.error(e);
                    return VFS.SQLITE_IOERR;
                }
            });
        }
        return VFS.SQLITE_OK;
    }
    xFileSize(fileId, pSize64) {
        const { file } = this.getFileStateByIdOrThrow(fileId);
        log(`xFileSize ${file.path}`);
        // console.log(file.block0.fileSize);
        pSize64.set(file.block0.fileSize);
        return VFS.SQLITE_OK;
    }
    xLock(fileId, flags) {
        const { file, db } = this.getFileStateByIdOrThrow(fileId);
        return this.handleAsync(async () => {
            log(`xLock ${file.path} ${flags}`);
            if (flags === VFS.SQLITE_LOCK_EXCLUSIVE) {
                this.pendingWrites.set(fileId, new Map());
            }
            try {
                const result = await this.webLocks.lock(file.path, flags);
                if (result === VFS.SQLITE_OK && flags === VFS.SQLITE_LOCK_SHARED) {
                    // console.log(file.block0);
                    // Update block 0 in case another connection changed it.
                    file.block0 = await db.run("readonly", ({ blocks }) => blocks.get(this.bound(0)));
                    // console.log(file.block0);
                }
                return result;
            }
            catch (e) {
                console.error(e);
                return VFS.SQLITE_IOERR;
            }
        });
    }
    xUnlock(fileId, flags) {
        const { file, db } = this.getFileStateByIdOrThrow(fileId);
        return this.handleAsync(async () => {
            log(`xUnlock ${file.path} ${flags}`);
            const filePendingWrites = this.pendingWrites.get(fileId);
            if (filePendingWrites && filePendingWrites.size > 0) {
                // console.time("writing bulk");
                // void db.run("readwrite", async ({ blocks }) => {
                //   for (const block of filePendingWrites.values()) {
                //     await blocks.put(block);
                //   }
                // });
                // await db.sync();
                const idb = await db.dbReady;
                await new Promise((resolve, reject) => {
                    // @ts-expect-error lib dom misses third argument
                    const tx = idb.transaction("blocks", "readwrite", {
                        durability: "relaxed",
                    });
                    const blocksStore = tx.objectStore("blocks");
                    for (const v of filePendingWrites.values()) {
                        blocksStore.put(v, v.offset);
                    }
                    tx.addEventListener("complete", () => {
                        resolve();
                    });
                    tx.addEventListener("abort", reject);
                    tx.addEventListener("error", reject);
                });
                // console.timeEnd("writing bulk");
            }
            this.pendingWrites.delete(fileId);
            try {
                await this.webLocks.unlock(file.path, flags);
                return VFS.SQLITE_OK;
            }
            catch (e) {
                console.error(e);
                return VFS.SQLITE_IOERR;
            }
        });
    }
    xSectorSize(fileId) {
        return 512;
    }
    xDeviceCharacteristics(fileId) {
        return (VFS.SQLITE_IOCAP_SAFE_APPEND |
            VFS.SQLITE_IOCAP_SEQUENTIAL |
            VFS.SQLITE_IOCAP_UNDELETABLE_WHEN_OPEN);
    }
    xAccess(name, flags, pResOut) {
        pResOut.set(0);
        return VFS.SQLITE_OK;
    }
    xDelete(name, syncDir) {
        const { db } = this.getFileStateByPathOrThrow(name);
        return this.handleAsync(async () => {
            new URL(name, "file://localhost/").pathname;
            try {
                const complete = db.run("readwrite", ({ blocks }) => {
                    return blocks.delete(this.bound(-Infinity));
                });
                if (syncDir)
                    await complete;
                return VFS.SQLITE_OK;
            }
            catch (e) {
                console.error(e);
                return VFS.SQLITE_IOERR;
            }
        });
    }
    txFileControl(fileId, op) {
        // console.debug("xFileControl", fileId, op);
        return VFS.SQLITE_NOTFOUND;
    }
    bound(begin, end = Infinity) {
        return IDBKeyRange.bound(begin, end);
    }
    getFileStateByPathOrThrow(path) {
        const dbName = path.replace(/-journal$/, "").replace(/-wal$/, "");
        const fileId = this.dbNameFileIdMap.get(dbName);
        if (!fileId) {
            throw new Error(`Failed to find db state for path: ${path}`);
        }
        return this.getFileStateByIdOrThrow(fileId);
    }
    getFileStateByIdOrThrow(fileId) {
        const state = this.filesState.get(fileId);
        if (!state)
            throw new Error("File not found");
        return state;
    }
}
function openDatabase(idbDatabaseName) {
    return new Promise((resolve, reject) => {
        const request = globalThis.indexedDB.open(`${idbDatabaseName}.sqlite`, 1);
        request.addEventListener("upgradeneeded", () => {
            request.result.createObjectStore("blocks");
        });
        request.addEventListener("success", () => {
            resolve(request.result);
        });
        request.addEventListener("error", () => {
            reject(request.error);
        });
    });
}
// read(position) {
//   let waitCursor = () => {
//     return new Promise((resolve, reject) => {
//       if (this.cursorPromise != null) {
//         throw new Error(
//           'waitCursor() called but something else is already waiting'
//         );
//       }
//       this.cursorPromise = { resolve, reject };
//     });
//   };
//   if (this.cursor) {
//     let cursor = this.cursor;
//     if (
//       cursor.direction === 'next' &&
//       position > cursor.key &&
//       position < cursor.key + 100
//     ) {
//       perf.record('stream-next');
//       cursor.advance(position - cursor.key);
//       return waitCursor();
//     } else if (
//       cursor.direction === 'prev' &&
//       position < cursor.key &&
//       position > cursor.key - 100
//     ) {
//       perf.record('stream-next');
//       cursor.advance(cursor.key - position);
//       return waitCursor();
//     } else {
//       // Ditch the cursor
//       this.cursor = null;
//       return this.read(position);
//     }
//   } else {
//     // We don't already have a cursor. We need to a fresh read;
//     // should we open a cursor or call `get`?
//     let dir = this.getReadDirection();
//     if (dir) {
//       // Open a cursor
//       this.prevReads = null;
//       let keyRange;
//       if (dir === 'prev') {
//         keyRange = IDBKeyRange.upperBound(position);
//       } else {
//         keyRange = IDBKeyRange.lowerBound(position);
//       }
//       let req = this.store.openCursor(keyRange, dir);
//       perf.record('stream');
//       req.onsuccess = (e) => {
//         perf.endRecording('stream');
//         perf.endRecording('stream-next');
//         let cursor = e.target.result;
//         this.cursor = cursor;
//         if (this.cursorPromise == null) {
//           throw new Error('Got data from cursor but nothing is waiting it');
//         }
//         this.cursorPromise.resolve(cursor ? cursor.value : null);
//         this.cursorPromise = null;
//       };
//       req.onerror = (e) => {
//         console.log('Cursor failure:', e);
//         if (this.cursorPromise == null) {
//           throw new Error('Got data from cursor but nothing is waiting it');
//         }
//         this.cursorPromise.reject(e);
//         this.cursorPromise = null;
//       };
//       return waitCursor();
//     } else {
//       if (this.prevReads == null) {
//         this.prevReads = [0, 0, 0];
//       }
//       this.prevReads.push(position);
//       this.prevReads.shift();
//       return this.get(position);
//     }
//   }
// }

const colors = ["yellow", "cyan", "magenta"];
const waSqliteWebBackend = ({ wasmUrl, pageSize, cacheSize, vfs: _vfs, }) => ({ dbName, stopped$ }) => {
    let sqlite3;
    let db;
    const vfs = _vfs ? _vfs : "minimal";
    let currentTransactionI = 0;
    let currentTransactionId;
    return {
        async initialize() {
            const module = await SQLiteAsyncModule({ locateFile: () => wasmUrl });
            sqlite3 = SQLite.Factory(module);
            const klass = vfs === "atomic"
                ? IDBAtomicVFS
                : vfs === "minimal"
                    ? IDCachedWritesVFS
                    : IDBBatchAtomicVFS;
            sqlite3.vfs_register(new klass(`idb-${vfs}-relaxed`, {
                purge: "manual",
                durability: "relaxed",
            }));
            db = await sqlite3.open_v2("wa-sqlite-" + dbName, undefined, `idb-${vfs}-relaxed`);
            await sqlite3.exec(db, `PRAGMA cache_size=${cacheSize === undefined ? -5000 : cacheSize};`);
            await sqlite3.exec(db, `PRAGMA page_size=${pageSize === undefined ? 32 * 1024 : pageSize};`);
            await sqlite3.exec(db, `PRAGMA journal_mode=MEMORY;`);
            // TODO: race condition on close may happen
            stopped$.subscribe(() => {
                if (sqlite3 && db !== undefined) {
                    void sqlite3.close(db);
                }
            });
        },
        async execQueries(queries, opts) {
            const logOpts = opts.log;
            if (!sqlite3 || db === undefined) {
                throw new Error("DB is not initialized");
            }
            const allResults = [];
            for (const q of queries) {
                const rows = [];
                const startTime = performance.now();
                const str = sqlite3.str_new(db, q.text);
                const prepare = await sqlite3.prepare_v2(db, sqlite3.str_value(str));
                if (!prepare) {
                    throw new Error(`Failed to prepare ${q.text} query`);
                }
                sqlite3.bind_collection(prepare.stmt, q.values);
                const columns = sqlite3.column_names(prepare.stmt);
                while ((await sqlite3.step(prepare.stmt)) === SQLite.SQLITE_ROW) {
                    if (columns.length > 0) {
                        rows.push(Object.fromEntries(sqlite3
                            .row(prepare.stmt)
                            .map((val, i) => [columns[i], val])));
                    }
                }
                sqlite3.str_finish(str);
                await sqlite3.finalize(prepare.stmt);
                allResults.push(rows);
                const end = performance.now();
                if (!opts.log.suppress) {
                    if (logOpts?.transactionId &&
                        logOpts.transactionId !== currentTransactionId &&
                        !logOpts.suppress) {
                        currentTransactionId = logOpts.transactionId;
                        currentTransactionI++;
                    }
                    if (!logOpts?.transactionId) {
                        currentTransactionId = undefined;
                    }
                    console.log(`%c[${dbName}]${logOpts?.transactionId
                        ? `[tr_id=${logOpts.transactionId.substring(0, 6)}]`
                        : ""} ${q.text} ${JSON.stringify(q.values)} Time: ${((end - startTime) /
                        1000).toFixed(4)}`, `color: ${currentTransactionI
                        ? colors[currentTransactionI % colors.length]
                        : "white"}`);
                }
            }
            return allResults;
        },
    };
};

export { waSqliteWebBackend };
